#/**
#*** Copyright (c) 1996 by Bill Bumgarner.  All Rights Reserved
#*** 
#*** This software may be used and distributed freely for any purpose
#*** provided that this notice is included unchanged on any and all copies.
#*** The author does not warrant or guarantee this software in any way.
#*** 
#*** 
#*** All queries, patches, and discussions of the PyObjC project should be
#*** directed to <objc-sig@python.org>.  The author can be reached via
#*** <bbum@friday.com>.
#*** 
#*** $RCSfile: PyServicesHandler.py,v $
#*** 
#*** $Revision: 1.3 $
#*** 
#*** $Date: 1996/10/24 23:18:24 $
#*** 
#**/

from Demiurge.DPyClassSupport import \
	 RestrictSetAttribute, UnrestrictSetAttribute, RestrictedSetAttr

import ObjC
import imp
import os
import posix
import sys
import traceback

HandlerError = 'HandlerError'
VisibleError = 'VisibleError'

def newHandler(servicesPaths):
   return PyServicesHandler(servicesPaths)

class PyServicesHandler:
   # This implementation of __setattr__ just calls RestrictedSetAttr.
   # Unfortunately, simply assigning __setattr__ to RestrictedSetAttr
   # in __init__ doesn't work (I don't know why).
   # 
   # RestrictedSetAttr() simply raises an exception if 'name' is
   # an instance variable that is not already defined within the
   # namespace of the instance self.  The functions
   # UnrestrictSetAttribute() and RestrictSetAttribute() are used
   # to turn on off the said restriction (see __init__()).
   # 
   # Using this, one can 'declare' the set of valid instance
   # variables by unrestricting __setattr__() in the beginning of
   # __init__(), initializing all instance variables to some known
   # value (None is valid), then restrictnig __setattr__() as the
   # last action of the __init__() method.
   # 
   # Subsequently, any attempts to set the value of any instance
   # variable that was not 'declared' as part of the object's
   # initialization will cause an AttributeError to be raised.
   # This *eliminates* mistyping instance variable names in any
   # assignments-- internally or externally-- and it tightly defines
   # the namespace used within an instance.   This is similar to
   # concretely defined API provided by, say, an Objective-C class
   # declaration.
   #
   # RestrictSetAttribute() and UnrestrictSetAttribute() work like
   # a stack-- every call to RestrictSetAttribute() should be
   # balanced by a call to UnrestrictSetAttribute().  Said 'restriction
   # scopes' can be nested.  [Ask me about a proper implementation of the
   # concept of 'super'].
   def __setattr__(self, name, value):
	  RestrictedSetAttr(self, name, value)

   def __init__(self, servicesPaths):
	  ### unrestrict addition of ivars
	  UnrestrictSetAttribute(self)

	  ### this is the variable 'declaration' described above.

	  self.daemon = None

	  ### filter services path
	  # expand paths with '~'
	  servicesPaths = map(os.path.expanduser, servicesPaths)

	  # only use paths that exist (and are directories)
	  servicesPaths = filter(os.path.isdir, servicesPaths)

	  # ensure that each path ends in '/'
	  self.servicesPaths = map(lambda p: ((p[-1:] == '/') and p) or (p + '/'),
							   servicesPaths)

	  # any paths left? :-)
	  if len(self.servicesPaths) == 0:
		 raise HandlerError,  'None of the supplied paths exists!'

	  ### initialize handler cache
	  self.clearCache()

	  ### restrict variable addition
	  RestrictSetAttribute(self)

   def clearCache(self):
	  self.handlerCache = {}

   def getHandlerForService(self, userD):
	  ## grab it from the cache
	  try:
		 return self.handlerCache[userD]
	  except:
		 pass

	  ## wasn't in the cache... search servicesPath
	  (head, tail) = os.path.split(userD)
	  module = None
	  for p in self.servicesPaths:
		 ## create test path
		 composite = p + head
		 if not (os.path.isdir(composite)):
			continue

		 ## directory exists-- find the module
		 try:
			(mF,mP,mD) = imp.find_module(tail, [composite])
		 except:
			print 'Failed to find {%s} in {%s}'%(tail, composite)
			continue

		 ## do the import (assumes description indicates python code)
		 try:
			module = imp.load_source(tail, mP, mF)
		 except:
			traceback.print_exc()
			continue

		 try:
			self.handlerCache[userD] = module.servicesHandler
			return module.servicesHandler
		 except:
			print 'Module {%s} does not contain ' \
				  + '"servicesHandler" function'%(composite)
			continue
	  ###
		 

   def serviceForwarder__withPasteboard__userData__(self,
													forwardingObject,
													pasteboard,
													userData):
	  ### dispatch the service
	  try:
		 ### find the implementor
		 h = self.getHandlerForService(userData)
		 
		 ### dispatch
		 if h == None:
			raise VisibleError, \
				  'Failed to find handler for service {%s}'%(userData)

		 h(self, pasteboard, userData)

	  ### return
	  except SystemExit:
		 print 'Daemon exiting'
	  except:
		 if sys.exc_type == VisibleError:
			## i tried except VisibleError:-- it didn't work
			forwardingObject.setErrorMessage__(sys.exc_value)
		 else:
			traceback.print_exc()

   def forkAndInvoke(self, func):
	  ### forks-- parent immediately returns.
	  ### child invokes method [useful for implementing services
	  ### that invoke other services, such as opening files].
	  pPID = os.getpid()
	  cPID = os.fork()
	  if cPID:
		 return ### parent returns

	  func(pPID)

	  ### if the func() happens to return, exit child without
	  ### raising an exception
	  posix._exit(0)
